# rSupLim is a function that draws one pseudo random of 
#     \sup_{m <= s <=1} \matchcal{H}^{m,+} (s) / g(s)
# by approximating the limiting distribution \matchcal{H}^{m,+} (s) using 
# functions of random walks as approximation for functionals of Brownian 
# motions. \matchcal{H}^{m,+} (s) is here based on PFM-OLS estimation.
# The function deals with intercept or linear time trend only as deterministic
# regressors but can easily be extended.

# Arguments
# trend   : logical, indicating whether the cointegrating regression contains 
#           only an intercept or a linear time trend as well
# k       : an integer, the number of I(1) regressors
# m       : a numeric, 0 < m < 1, the length of the calibration period as a 
#           fraction of the sample size
# N       : an integer, the number of cointegrating relationships
# bm_prec : an integer, the number of iid. standard normal random variables 
#           used for a random walk that approximates a Brownian motion

# Value
# A named numeric vector containing one pseudo random number generated for
# each of the three PFM-OLS based monitoring test statistics.

# Author
# Etienne Theising

rSupLim = function(trend, k, m, N, bm_prec = 1000L) {
  m.idx = floor(m * bm_prec)
  p = 1
  Dt = matrix(1, nrow = 1, ncol = bm_prec)
  Dt_int = seq(1 / bm_prec, 1, by = 1 / bm_prec)
  if (isTRUE(trend)) {
    p = 2
    Dt = rbind(Dt, seq(1 / bm_prec, 1, by = 1 / bm_prec))
    Dt_int = rbind(Dt_int, 0.5 * Dt_int^2)
  }
  W_uv = t(apply(replicate(N, rnorm(bm_prec, sd = sqrt(1 / bm_prec))), 
                 2, cumsum))
  W_v = t(apply(replicate(k * N, rnorm(bm_prec, sd = sqrt(1 / bm_prec))), 
                2, cumsum))
  # Calculate the matrix factors on the right hand side of \hat{W}_{u\cdot v,n}
  sum_JJ_int = matrix(0, nrow = p + k, ncol = p + k)
  sum_JdW_int = matrix(0, nrow = p + k, ncol = 1)
  for (j in 1:N) {
    J_j = rbind(Dt[, 1:m.idx], W_v[((j - 1) * k + 1):(j * k), 1:m.idx])
    JJ_int = matrix(apply(apply(J_j, 2, function(x) outer(X = x, Y = x)), 
                          1, sum), ncol = p + k) / bm_prec
    JdW_int = matrix(c(W_uv[j, 1], rep(0, p + k - 1)), ncol = 1) + 
      apply(rbind(Dt[, 1:(m.idx - 1)], 
                  W_v[((j - 1) * k + 1):(j * k), 1:(m.idx - 1)]) %*% 
              diag(W_uv[j, 2:m.idx] - W_uv[j, 1:(m.idx - 1)]), 1, sum)
    sum_JJ_int = sum_JJ_int + JJ_int
    sum_JdW_int = sum_JdW_int + JdW_int
  }
  cons_factor = solve(sum_JJ_int) %*% sum_JdW_int
  rm(sum_JJ_int, sum_JdW_int, JJ_int, JdW_int)
  
  W_v_int = t(apply(W_v, 1, cumsum)) / bm_prec
  # calculate \hat{W}_{u\cdot v,n} for every l with m <= l / bm_prec <= 1
  W_uv_hat = matrix(0, nrow = N, ncol = bm_prec)
  for (j in 1:N) {
    W_uv_hat[j, ] = W_uv[j, ] - 
      as.numeric(t(rbind(Dt_int, W_v_int[1:k + (j - 1) * k, ])) %*% cons_factor)
  }
  rm(cons_factor, W_uv, W_v, Dt_int, W_v_int)
  
  W_uv_hat0m = matrix(W_uv_hat[, 1:m.idx], nrow = N)
  W_uv_hat = matrix(W_uv_hat[, m.idx:bm_prec], nrow = N)
  H1_PFM = apply((W_uv_hat - W_uv_hat[, 1])^2, 2, sum) / 
    sum(W_uv_hat[, 1]^2)
  # note that integral approximation and sum_n_to_N are just one sum for H2_PFM
  H2_PFM = apply(t(apply(W_uv_hat^2 / bm_prec, 1, cumsum)), 2, sum) / 
                sum(W_uv_hat0m^2 / bm_prec)
  int_W_uv_hat = t(apply(W_uv_hat / bm_prec, 1, cumsum))
  int_W_uv_hat0m = apply(W_uv_hat0m / bm_prec, 1, sum)
  H3_PFM = apply(int_W_uv_hat^2, 2, sum) / sum(int_W_uv_hat0m^2)
  rm(W_uv_hat, int_W_uv_hat0m)
  grid = seq(m.idx / bm_prec, 1, by = 1 / bm_prec)
  return (c(H1.PFM = max(H1_PFM / grid^(2 * p)), 
            H2.PFM = max(H2_PFM / grid^(2 * p + 1)), 
            H3.PFM = max(H3_PFM / grid^(2 * p + 2))
            ))
}
